# -*- coding: utf-8 -*-
"""
   Wiki+

   Copyright (C) 2005 Yong-Jin Cho <yongjin.cho@gmail.com>
   Wiki+ is distributed under GNU GPL. See COPYING for details.

   This program is based on PikiPiki which is a simple wiki engine.
   PikiPiki project homepage is http://sourcefrog.net/projects/piki.

   For more information, visit http://wikiplus.kldp.net.
"""
# Imports -----------------------------------------------------------
import cgi, sys, string, os, fcntl, re, errno, time, urllib
import config



# Exceptions --------------------------------------------------------

# for page class
PageError = 'PageError'
NoPage = 'NoPage'
EmptyText = 'EmptyText'
UnchangedText = 'UnchangedText'
UpdateConflict = 'UpdateConflict'

# for action handler
UnknownAction = 'UnknownAction'

# for macro expander
UnexpectedToken = 'UnexpectedToken'



# Page class --------------------------------------------------------

class Page:
    """Page management class

       Directory Structure
       -------------------
        data/ --+-- text/ ---+-- pageid1
                |            `-- pageid2
                |
                +-- backup/ -+-- pageid1            --> editlog
                |            +-- pageid1.timestamp
                |            +-- pageid1.timestamp
                |            +-- pageid2
                |            `-- pageid2.timestamp
                |
                `-- editlog

       Page ID
       -------
        quoted page name

       Editlog File Structure
       ----------------------
        +-------------------------------------------------------+
        |page id<TAB>modified time<TAB>host<TAB>comment<NEWLINE>|
        |    .             .             .         .            |
        |                                                       |
    """

    def __init__(self, name, timestamp=None):
        self.name = name
        self.url_name = urlencode(name)
        self.file_name = self.url_name.replace('%', '_')
        self.timestamp = timestamp


    def _text_path(self):
        if not self.timestamp:
            return os.path.join(config.text_dir, self.file_name)
        else:
            return os.path.join(config.backup_dir,
                                        self.file_name + '.' + self.timestamp)

    def _log_path(self):
        return os.path.join(config.backup_dir, self.file_name)

    def _tmp_path(self):
        return os.path.join(config.data_dir,
                        ('#' + self.file_name + '.' + str(os.getpid()) + '#'))

    def _check_current(self):
        if self.timestamp:
            raise PageError


    def exists(self):
        return os.path.exists(self._text_path())


    def link(self):
        self._check_current()
        if self.exists():
            return link_tag(self.url_name, self.name)
        else:
            return link_tag(self.url_name, '?') + self.name


    def get_mtime(self):
        try:
            return os.path.getmtime(self._text_path())
        except OSError, er:
            if er.errno == errno.ENOENT:
                raise NoPage
            else:
                raise er
        

    def get_size(self):
        try:
            return os.path.getsize(self._text_path())
        except OSError, er:
            if er.errno == errno.ENOENT:
                raise NoPage
            else:
                raise er


    def get_text(self):
        try:
            text_file = open(self._text_path(), 'rb')
        except IOError, er:
            if er.errno == errno.ENOENT:
                raise NoPage
            else:
                raise er
        try:
            text = text_file.read()
        finally:
            text_file.close()
        return text


    def update_text(self, timestamp, newtext, comment):
        self._check_current()
        newtext = newtext.replace('\r', '')
        if not newtext:
            raise EmptyText

        tmp_path = self._tmp_path()
        tmp_file = open(tmp_path, 'wb')
        tmp_file.write(newtext)
        tmp_file.close()

        text_path = self._text_path()
        if self.exists():
            mtime = self.get_mtime()
            if timestamp != str(mtime):
                raise UpdateConflict
            elif newtext == self.get_text():
                raise UnchangedText
            os.rename(text_path, os.path.join(config.backup_dir,
                                            self.file_name + '.' + str(mtime)))

        os.chmod(tmp_path, 0666)
        os.rename(tmp_path, text_path)

        timestamp = str(self.get_mtime())
        host = os.environ.get('REMOTE_ADDR', '')
        editlog_add(self._log_path(), self.file_name, timestamp, host, comment)
        os.chmod(self._log_path(), 0666)
        editlog_add(config.editlog_path,
                                      self.file_name, timestamp, host, comment)


    def get_backup_list(self):
        tmp_list = os.listdir(config.backup_dir)
        prefix = self.file_name + '.'
        tmp_list = filter(lambda x: x.startswith(prefix), tmp_list)
        return map(lambda x: x.split('.')[1], tmp_list)


    def send_page(self, msg=None):
        link = '%s?action=FullSearch&amp;value=%s' % \
                                                (self.url_name, self.url_name)
        send_title(self.name, link, msg)

        try:
            text = self.get_text()
        except NoPage:
            print link_tag(self.url_name + '?action=edit', 'Create this page')
            last_modified = None
        else:
            Formatter().format(text)
            mtime = self.get_mtime()
            last_modified = time.strftime(config.datetime_fmt, 
                                            time.localtime(self.get_mtime()))

        send_footer(self.name, True, last_modified)


    def send_editor(self, preview=False, **kw):
        self._check_current()
        if not preview:
            send_title('Edit "%s"' % self.name)
            if self.exists():
                timestamp = str(self.get_mtime())
                newtext = self.get_text()
            else:
                timestamp = ''
                newtext = 'Dscribe %s here.' % self.name
            comment = ''

        else:
            send_title('Preview of "%s"' % self.name)
            timestamp = kw['timestamp']
            newtext = kw['newtext'].replace('\r', '')
            comment = kw['comment']

        print '<form method="post" action="%s/%s#preview">' % \
              (get_scriptname(), self.url_name)
        print '<input type="hidden" name="action" value="update" />'
        print '<input type="hidden" name="timestamp" value="%s" />' % timestamp
        print '<textarea name="newtext" rows="18" cols="90" ' \
              'style="width:95%%;word-wrap:virtual;">%s</textarea><br />' % newtext
        print 'Comment: <input type="text" name="comment" value="%s" ' \
              'size="50" /><br />' % comment
        print '<input type="submit" name="save" value="Save" /> ' \
              '<input type="submit" name="preview" value="Preview" /></form>'

        if preview:
            print '<a name="preview"></a><hr />'
            Formatter().format(newtext)

        send_footer(self.name, False)


def get_page_list():
    tmp_list = map(lambda x:x.replace('_', '%'), os.listdir(config.text_dir))
    return map(urldecode, tmp_list)



# Editlog -----------------------------------------------------------

# Functions to keep track of when people have changed pages, so we can
# do the recent changes page and so on.
# The editlog is stored with one record per line, as tab-separated
# words: pageid, timestamp, host, comment

def editlog_add(log_path, pageid, timestamp, host, comment):
    entry = string.join((pageid, timestamp, host, comment), "\t") + "\n"
    editlog = open(log_path, 'ab')
    try: 
        fcntl.flock(editlog.fileno(), fcntl.LOCK_EX)
        editlog.seek(0, 2)                  # to end
        editlog.write(entry)
    finally:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()


def editlog_raw_lines(log_path):
    editlog = open(log_path, 'rb')
    try:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_SH)
        return editlog.readlines()
    finally:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()



# Formatter ---------------------------------------------------------

class Formatter:
    def __init__(self):
        self._line_re = re.compile(r'^(?:' \
            r'(?P<heading>(?P<hmarker>={1,6})\s.*\s(?P=hmarker))|' \
            r'(?P<rule>-{4,})|' \
            r'(?P<list>\s+(?:(?P<ol>\d+\.\s)|(?P<ul>\*\s)).*)|' \
            r'(?P<normal>.+)|' \
            r'(?P<blank>)' \
            r')$')
        self._indent_re = re.compile(r'\s+')
        self._inline_re = re.compile(
            r"(?P<emph>'{2,3})|" \
            r'(?P<word>\b(?:[A-Z][a-z]+){2,}\b)|' \
            r'(?P<eword>\[[^\[\]]+\])|' \
            r'(?P<url>(?:http|ftp|nntp|news|mailto)\:[^\s\'"]+\S)|' \
            r'(?P<email>[-\w._+]+\@[\w.-]+)|' \
            r'(?P<macro>\[\[[A-Za-z][A-Za-z0-9]*(?:\(.*\))?\]\])|' \
            r'(?P<code>\{\{\{[^(?:\}\}\})]+\}\}\})|' \
            r'(?P<pre>(?:\{\{\{|\}\}\}))|' \
            r'(?P<ent>[<>&])')

        self._list_stack = [(0, None)]      # (level, order)
        self._in_pre = False
        self._is_em = self._is_b = False


    def format(self, raw):
        raw_lines = raw.expandtabs().split('\n')
        for raw_line in raw_lines:
            if not self._in_pre:
                match = self._line_re.match(raw_line)
                for type, s in match.groupdict().items():
                    if s != None and hasattr(self, '_' + type + '_line'):
                        apply(getattr(self, '_' + type + '_line'), (match,))
                        break
            else:
                print self._inline(raw_line)

        while self._list_stack[-1][0]:
            prev_order = self._list_stack.pop()[1]
            print '</li></%s>' % prev_order


    # line formatting part
    def _normal_line(self, match):
        print self._inline(match.group(0))
        if config.auto_linebreak and not self._in_pre:
            print '<br />'


    def _list_line(self, match):
        if match.group('ul'):
            order = 'ul'
        else:
            order = 'ol'
        level = len(self._indent_re.match(match.group(0)).group(0))
        prev_level = self._list_stack[-1][0]

        if level == prev_level:
            print '</li><li>'
        elif level > prev_level:
            print '<%s><li>' % order
            self._list_stack.append((level, order))
        else:
            while level < prev_level:
                prev_order = self._list_stack.pop()[1]
                print '</li></%s>' % prev_order
                prev_level = self._list_stack[-1][0]
            print '</li><li>'

        print self._inline(match.group(0)[level + len(match.group(order)):])


    def _heading_line(self, match):
        level = len(match.group('hmarker'))
        print '<h%d>' % level
        print self._inline(match.group(0)[level + 1: -(level + 1)])
        print '</h%d>' % level


    def _rule_line(self, match):
        if len(match.group(0)) <= 4:
            print '<hr />'
        else:
            print '<hr size="%d" />' % (len(match.group(0)) - 2)


    def _blank_line(self, match):
        while self._list_stack[-1][0]:
            prev_order = self._list_stack.pop()[1]
            print '</li></%s>' % prev_order
        else:
            print '<p></p>'


    # inline replacement part
    def _inline(self, line):
        return self._inline_re.sub(self._replace, line)


    def _replace(self, match):
        for type, hit in match.groupdict().items():
            if hit:
                return apply(getattr(self, '_' + type + '_repl'), (hit,))


    def _emph_repl(self, word):
        if not self._in_pre:
            if len(word) == 3:
                self._is_b = not self._is_b
                return ['</b>', '<b>'][self._is_b]
            else:
                self._is_em = not self._is_em
                return ['</em>', '<em>'][self._is_em]
        else:
            return word

    def _word_repl(self, word):
        if not self._in_pre:
            return Page(word).link()
        else:
            return word

    def _eword_repl(self, word):
        if not self._in_pre:
            return Page(word[1:-1]).link()
        else:
            return word

    def _url_repl(self, word):
        if not self._in_pre:
            return '<a href="%s">%s</a>' % (word, word)
        else:
            return word

    def _email_repl(self, word):
        if not self._in_pre:
            return '<a href="mailto:%s">%s</a>' % (word, word)
        else:
            return word

    def _macro_repl(self, word):
        if not self._in_pre:
            pos = word.find('(')
            if pos == -1:
                name = word[2:-2]
                arg = ''
            else:
                name = word[2:pos]
                arg = word[pos+1:-3]
            return expand_macro(name, arg)
        else:
            return word

    def _code_repl(self, word):
        if not self._in_pre:
            return '<code>%s</code>' % word[3:-3]
        else:
            return word

    def _pre_repl(self, word):
        if not self._in_pre and word == '{{{':
            self._in_pre = True
            return '<pre>'
        elif self._in_pre and word == '}}}':
            self._in_pre = False
            return '</pre>'
        else:
            return word

    def _ent_repl(self, s):
        return {'&': '&amp;', '<': '&lt;', '>': '&gt;'}[s]



# Macro functions ---------------------------------------------------

class Macro:
    def __init__(self):
        self._macros = {}


    def expand(self, name):
        if not self._macros.has_key(name):
            try:
                module = __import__('macro.' + name, {}, {}, ['macro'])
            except ImportError:
                return '<i>"%s" macro does not exist.</i>' % name
            else:
                self._macros[name] = module.expand

        return self._macros[name](self)


def init_macro(page_name, form):
    global _macro
    _macro = Macro()
    _macro.page_name = page_name
    _macro.form = form


def expand_macro(name, arg):
    global _macro
    _macro.arg = arg
    return _macro.expand(name)



# Action handlers ---------------------------------------------------

def do_show(page_name, form):
    pg = Page(page_name, form.getvalue('timestamp'))
    if not pg.timestamp:
        msg = None
    else:
        try:
            mtime = int(pg.timestamp)
        except:
            mtime = 0
        date = time.strftime(config.datetime_fmt, time.localtime(mtime))
        if pg.exists():
            msg = '<b>Revision as of %s</b>' % date
        else:
            pg = Page(page_name)
            msg = '<b>No revision dated %s</b>' % date
    pg.send_page(msg)


def do_edit(page_name, form):
    Page(page_name).send_editor()


def do_update(page_name, form):
    timestamp = form.getvalue('timestamp')
    newtext = form.getvalue('newtext', '')
    comment = form.getvalue('comment', '')

    pg = Page(page_name)
    if form.getvalue('save'):
        try:
            pg.update_text(timestamp, newtext, comment)
        except EmptyText:
            msg = '<b>You cannot save empty pages.</b>'
        except UnchangedText:
            msg = '<b>You cannot save unchanged pages.</b>'
        except UpdateConflict:
            msg = '<p><b>Sorry, someone else saved the page while you edited' \
                  ' it.</b></p><p>Please do the following: Cut&amp;paste you' \
                  'r changes from the following box. Then click EditPage aga' \
                  'in(or shift-click it to open in a new window). Now re-add' \
                  ' your changes to the current page contents.</p><p><em>Do ' \
                  'not just replace the content editbox with your version of' \
                  ' the page, because that would delete the changes of the o' \
                  'ther person, which is excessively rude!</em></p><p>Follow' \
                  'ing is the page you edited.</p>'
            msg += '<textarea wrap="virtual" rows="17" cols="80">%s' \
                   '</textarea>' % newtext
        else:
            msg = '<b>Thankyou for your changes. Your attention to detail is' \
                  'appreciated.</b>'
        pg.send_page(msg)

    else:
        pg.send_editor(True,
                        timestamp=timestamp, newtext=newtext, comment=comment)


def do_info(page_name, form):
    def print_info_row(pg):
        mtime = pg.get_mtime()
        print '<tr><td>%s</td>' % \
                    time.strftime(config.datetime_fmt, time.localtime(mtime))
        print '<td align="right">%d</td><td>' % pg.get_size()
        if not pg.timestamp:
            url_suffix = ''
        else:
            url_suffix = '?timestamp=' + pg.timestamp
        print link_tag(pg.url_name + url_suffix, 'view')
        print '</td></tr>'

    pg = Page(page_name)
    backup_list = pg.get_backup_list()
    backup_list.sort()
    backup_list.reverse()

    if len(backup_list):
        send_title('Info of "' + page_name + '"')

        print '<table><tr><th>Date</th><th>Size</th><th>Actions</th></tr>'
        if pg.exists():
            print_info_row(pg)
        for timestamp in backup_list:
            pg = Page(page_name, timestamp)
            print_info_row(pg)
        print '</table>'

        send_footer(page_name, False)
    else:
        pg.send_page('<b>No older revision exists.</b>')


def get_handler(action):
    basic_handlers = {
        'show':     do_show,
        'edit':     do_edit,
        'update':   do_update,
        'info':     do_info
    }
    if basic_handlers.has_key(action):
        return basic_handlers[action]

    try:
        module = __import__('action.' + action, {}, {}, ['action'])
    except ImportError:
        raise UnknownAction, action
    else:
        return module.do



# Utility functions -------------------------------------------------

def emit_header():
    print "Content-type: text/html"
    print


def urlencode(s):
    t = urllib.quote(s, '')
    t = t.replace('-', '%2D')
    t = t.replace('.', '%2E')
    return t.replace('_', '%5F')


def urldecode(s):
    return urllib.unquote(s)


def quote_name(name):
    return urlencode(name).replace('%', '_')


def unquote_name(name):
    return urldecode(name.replace('_', '%'))


def get_scriptname():
    return os.environ.get('SCRIPT_NAME', '')


def get_basename(path):
    import posixpath, dospath, macpath

    for pathmodule in (posixpath, dospath, macpath):
        basename = pathmodule.basename(path)
        if basename != path:
            return basename
    return path


def link_tag(link, text=None):
    if not text:
        text = link
    return '<a href="%s/%s">%s</a>' % (get_scriptname(), link, text)


def send_title(title, link=None, msg=None):
    print '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" ' \
          '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
    print '<html><head><title>%s</title>' % title
    print '<meta http-equiv="content-type" content="text/html;' \
          'charset=%s" />' % config.charset
    print '<link rel="stylesheet" type="text/css" href="%s" /></head>' \
          '<body><br />' % config.css_url
    print link_tag(urlencode(config.front_page),
                  '<img class="logo" src="%s" alt="logo" />' % config.logo_url)
    print '<div class="title">'
    if link:
        print link_tag(link, title)
    else:
        print title
    print '</div>'
    print link_tag(urlencode(config.recent_changes), config.recent_changes)
    print ' | '
    print link_tag(urlencode(config.find_page), config.find_page)
    print ' | '
    print link_tag(urlencode(config.title_index), config.title_index)
    print '<hr /><div class="body">'
    if msg:
        print '%s<hr />' % msg


def send_footer(page_name, editable=True, mod_string=None):
    print '</div><hr /><div class="footmenu">'
    if editable:
        print link_tag(urlencode(page_name) + '?action=edit', config.edit_page)
    else:
        print link_tag(urlencode(page_name), config.show_page)
    if mod_string:
        print ' (last modified %s)' % mod_string
    print '</div><div class="power">Powered by ' \
          '<a href="http://wikiplus.kldp.net">Wiki+</a><br />' \
          '<a href="http://validator.w3.org/check/referer">XHTML 1.0</a> | ' \
          '<a href="http://jigsaw.w3.org/css-validator/check/referer">' \
          'CSS 2.0</a></div><br /><br /><br /></body></html>'



# CGI main ----------------------------------------------------------

def run():
    try:
        emit_header()

        # get request
        path_info = os.environ.get('PATH_INFO', '')
        if len(path_info) > 1:
            page_name = path_info[1:]
        else:
            page_name = config.front_page
        form = cgi.FieldStorage()
        action = form.getvalue('action', 'show')

        # inititalize macro module
        init_macro(page_name, form)

        # handle request
        handler = get_handler(action)
        handler(page_name, form)

    except UnknownAction, action:
        print 'Unknown action: ' + action

    except:
        cgi.print_exception()

    sys.stdout.flush()
